/* TODO: entity start/end markers, xml:base */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "system.h"
#include "charset.h"
#include "string16.h"
#include "dtd.h"
#include "input.h"
#include "xmlparser.h"
#include "stdio16.h"
#include "version.h"
#include "namespaces.h"
#include "infoset-print.h"

#define INDENT_STEP 2

static void children(FILE16 *f, int level, Dtd dtd, XBit *bits, int nbits, 
		     int *ecounter, int *acounter);
static void notations(FILE16 *f, int level, Dtd dtd);
static void entities(FILE16 *f, int level, Dtd dtd, Entity docent);
static void baseURI(FILE16 *f, int level, const char8 *uri);
static void standalone(FILE16 *f, int level, StandaloneDeclaration sd);
static void version(FILE16 *f, int level, const char8 *version);
static void item(FILE16 *f, int level, Dtd dtd, XBit bit,
		 int *ecounter, int *acounter);
static void xdtd(FILE16 *f, int level, Dtd dtd, XBit bit);
static void element(FILE16 *f, int level, Dtd dtd, XBit bit,
		    int *ecounter, int *acounter);
static void pi(FILE16 *f, int level, Dtd dtd, XBit bit);
static void cdsect(FILE16 *f, int level, Dtd dtd, XBit bit);
static void pcdata(FILE16 *f, int level, Dtd dtd, XBit bit);
static void comment(FILE16 *f, int level, Dtd dtd, XBit bit);
static void name(FILE16 *f, int level, 
		 const char8 *uri, const Char *local, const Char *prefix);
static void document_element(FILE16 *f, int level, XBit bit);
static void character(FILE16 *f, int level, Dtd dtd, char *ecw, int c);
static void attributes(FILE16 *f, int level, Dtd dtd, XBit bit, int *acounter);
static void attribute(FILE16 *f, int level, Attribute a);
static void namespace_attributes(FILE16 *f, int level, Dtd dtd, XBit bit, 
				 int *acounter);
static void inscope_namespaces(FILE16 *f, int level, Dtd dtd, XBit bit);
static void internal_entity(FILE16 *f, int level, Entity entity);
static void external_entity(FILE16 *f, int level, Entity entity);
static void unparsed_entity(FILE16 *f, int level, Entity entity);

static void simple(FILE16 *f, int level, char *name, const char *value);
static void Simple(FILE16 *f, int level, char *name, const Char *value);
static void pointer(FILE16 *f, int level, char *name, Char *id);
static Char *make_id(const char *type, const Char *name, int count);
static void indent(FILE16 *f, int level);

static void escape(FILE16 *f, const char8 *text);
static void Escape(FILE16 *f, const Char *text);
static void escape1(FILE16 *f, int c);

void infoset_print(FILE16 *f, Parser p, XBit *bits, int nbits)
{
    int i, e=1, a=1;
    Dtd dtd = p->dtd;

    Fprintf(f, "<document xmlns:xsi=\"http://www.w3.org/2000/10/XMLSchema-instance\"\n          xmlns=\"http://www.w3.org/2000/12/XMLInfoset\">\n");

    children(f, 1, dtd, bits, nbits, &e, &a);

    pointer(f, 1, "documentEntity", strdup_char8_to_Char("docent"));

    for(i=0; i<nbits; i++)
	if(bits[i]->type == XBIT_start || bits[i]->type == XBIT_empty)
	    document_element(f, 1, bits[i]);

    notations(f, 1, dtd);

    entities(f, 1, dtd, p->document_entity);

    baseURI(f, 1, EntityBaseURL(p->document_entity));

    standalone(f, 1, p->standalone);

    version(f, 1, p->document_entity->version_decl);

    simple(f, 1, "allDeclarationsProcessed", "true");

    Fprintf(f, "</document>\n");
}

static void children(FILE16 *f, int level, Dtd dtd, XBit *bits, int nbits,
		     int *ecounter, int *acounter)
{
    int i;

    if(nbits == 0)
    {
	indent(f, level);
	Fprintf(f, "<children/>\n");
    }
    else
    {	
	indent(f, level);
	Fprintf(f, "<children>\n");
	for(i=0; i<nbits; i++)
	    item(f, level+1, dtd, bits[i], ecounter, acounter);
	indent(f, level);
	Fprintf(f, "</children>\n");
    }
}

static void document_element(FILE16 *f, int level, XBit bit)
{
    Char el[] = {'r','o','o','t','e','l','e','m','e','n','t',0};

    pointer(f, level, "documentElement", el);
}

static void notations(FILE16 *f, int level, Dtd dtd)
{
    NotationDefinition n;
    Char *id;

    indent(f, level);
    Fprintf(f, "<notations>\n");

    for(n=NextNotationDefinition(dtd, 0); n; n=NextNotationDefinition(dtd, n))
    {
	if(n->tentative)
	    continue;

	id = make_id("notation", n->name, 0);

	indent(f, level+1);
	Fprintf(f, "<notation id=\"%S\">\n", id);
	free(id);

	Simple(f, level+2, "name", n->name);

	simple(f, level+2, "systemIdentifier", n->systemid);

	simple(f, level+2, "publicIdentifier", n->publicid);

	indent(f, level+1);
	Fprintf(f, "</notation>\n");
    }

    indent(f, level);
    Fprintf(f, "</notations>\n");
}

static void entities(FILE16 *f, int level, Dtd dtd, Entity docent)
{
    Entity e;

    indent(f, level);
    Fprintf(f, "<entities>\n");

    external_entity(f, level+1, docent);

    for(e=dtd->predefined_entities; e; e=e->next)
	internal_entity(f, level+1, e);

    for(e=NextEntity(dtd, 0); e; e=NextEntity(dtd, e))
	if(e->type == ET_internal)
	    internal_entity(f, level+1, e);
	else if(e->notation)
	{
	    unparsed_entity(f, level+1, e);
	}
	else
	{
	    external_entity(f, level+1, e);
	}

    indent(f, level);
    Fprintf(f, "</entities>\n");
}

static void internal_entity(FILE16 *f, int level, Entity entity)
{
    Char *id = make_id("entity", entity->name, 0);

    indent(f, level);
    Fprintf(f, "<internalEntity id=\"%S\">\n", id);
    free(id);

    Simple(f, level+1, "name", entity->name);

    Simple(f, level+1, "content", entity->text);

    indent(f, level);
    Fprintf(f, "</internalEntity>\n");
}

static void external_entity(FILE16 *f, int level, Entity entity)
{
    Char docent[] = {'d', 'o', 'c', 'e', 'n', 't', 0};
    Char *id = entity->name ? make_id("entity", entity->name, 0) : docent;

    indent(f, level);
    Fprintf(f, "<externalEntity id=\"%S\">\n", id);
    free(id);

    Simple(f, level+1, "name", entity->name);

    simple(f, level+1, "systemIdentifier", entity->systemid);

    simple(f, level+1, "publicIdentifier", entity->publicid);

    simple(f, level+1, "baseURI", EntityBaseURL(entity));

    simple(f, level+1, "charset", CharacterEncodingName[entity->encoding]);
    
    indent(f, level);
    Fprintf(f, "</externalEntity>\n");
}

static void unparsed_entity(FILE16 *f, int level, Entity entity)
{
    Char *id = make_id("entity", entity->name, 0);

    indent(f, level);
    Fprintf(f, "<unparsedEntity id=\"%S\">\n", id);
    free(id);

    Simple(f, level+1, "name", entity->name);

    simple(f, level+1, "systemIdentifier", entity->systemid);

    simple(f, level+1, "publicIdentifier", entity->publicid);

    id = make_id("notation", entity->notation->name, 0);
    pointer(f, level+1, "notation", id);
    free(id);

    indent(f, level);
    Fprintf(f, "</unparsedEntity>\n");
}

static void baseURI(FILE16 *f, int level, const char8 *uri)
{
    simple(f, level, "baseURI", uri);
}

static void standalone(FILE16 *f, int level, StandaloneDeclaration sd)
{
    simple(f, level, "standalone",
	   sd == SDD_unspecified ? 0 : StandaloneDeclarationName[sd]);
}

static void version(FILE16 *f, int level, const char8 *version)
{
    simple(f, level, "version", version);
}

static void simple(FILE16 *f, int level, char *name, const char *value)
{
    indent(f, level);

    if(value)
    {
	Fprintf(f, "<%s>", name);
	escape(f, value);
	Fprintf(f, "</%s>\n", name);
    }
    else
	Fprintf(f, "<%s xsi:null=\"true\"/>\n", name);
}

static void Simple(FILE16 *f, int level, char *name, const Char *value)
{
    indent(f, level);

    if(value)
    {
	Fprintf(f, "<%s>", name);
	Escape(f, value);
	Fprintf(f, "</%s>\n", name);
    }
    else
	Fprintf(f, "<%s xsi:null=\"true\"/>\n", name);
}

static void pointer(FILE16 *f, int level, char *name, Char *id)
{
    indent(f, level);

    Fprintf(f, "<%s><pointer ref=\"%S\"/></%s>\n", name, id, name);
}

static void indent(FILE16 *f, int level)
{
    int i;

    for(i=0; i<level * INDENT_STEP; i++)
	Fprintf(f, " ");
}

static void item(FILE16 *f, int level, Dtd dtd, XBit bit,
		 int *ecounter, int *acounter)

{
    switch(bit->type)
    {
	case XBIT_dtd:
	    xdtd(f, level, dtd, bit);
	    break;
	case XBIT_start:
	    element(f, level, dtd, bit, ecounter, acounter);
	    break;
	case XBIT_empty:
	    element(f, level, dtd, bit, ecounter, acounter);
	    break;
	case XBIT_pi:
	    pi(f, level, dtd, bit);
	    break;
	case XBIT_cdsect:
	    cdsect(f, level, dtd, bit);
	    break;
	case XBIT_pcdata:
	    pcdata(f, level, dtd, bit);
	    break;
	case XBIT_comment:
	    comment(f, level, dtd, bit);
	    break;
	default:
	    fprintf(stderr, "***%s\n", XBitTypeName[bit->type]);
	    exit(1);
	    break;
    }
}

static void xdtd(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    indent(f, level);
    Fprintf(f, "<documentTypeDeclaration>\n");

    simple(f, level+1, "systemIdentifier", 
	   dtd->external_part ? dtd->external_part->systemid : 0);

    simple(f, level+1, "publicIdentifier", 
	   dtd->external_part ? dtd->external_part->publicid : 0);

    children(f, level+1, dtd, bit->children, bit->nchildren, 0, 0);

    indent(f, level);
    Fprintf(f, "</documentTypeDeclaration>\n");
}

static void element(FILE16 *f, int level, Dtd dtd, XBit bit,
		     int *ecounter, int *acounter)
{
    static Char empty[] = {0};

    indent(f, level);
    if(!bit->parent)
	Fprintf(f, "<element id=\"rootelement\">\n");
    else
	Fprintf(f, "<element>\n");


    name(f, level+1,
	 bit->ns_element_definition ? 
  	   bit->ns_element_definition->namespace->uri : 0,
	 bit->element_definition->local,
	 bit->element_definition->prefix ?
	   bit->element_definition->prefix : empty);

    children(f, level+1, dtd, bit->children, bit->nchildren,
	     ecounter, acounter);

    attributes(f, level+1, dtd, bit, acounter);

    namespace_attributes(f, level+1, dtd, bit, acounter);

    inscope_namespaces(f, level+1, dtd, bit);

    baseURI(f, 1, EntityBaseURL(bit->entity)); /* XXX xml:base */

    indent(f, level);
    Fprintf(f, "</element>\n");
}

static void pi(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    indent(f, level);
    Fprintf(f, "<processingInstruction>\n");

    Simple(f, level+1, "target", bit->pi_name);

    Simple(f, level+1, "content", bit->pi_chars);

    baseURI(f, level+1, EntityBaseURL(bit->entity)); /* XXX xml:base */

    indent(f, level);
    Fprintf(f, "</processingInstruction>\n");
}

static void cdsect(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    Char *p;

    indent(f, level);
    Fprintf(f, "<cDATAStartMarker/>\n");

    for(p=bit->pcdata_chars; *p; p++)
	character(f, level, dtd, "false", *p);

    indent(f, level);
    Fprintf(f, "<cDATAEndMarker/>\n");
}

static void pcdata(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    Char *p;
    char *ecw;

    ecw = !bit->parent->element_definition->declared ? 0 :
	    bit->parent->element_definition->type == CT_element ? "true" :
	                                                          "false";

    for(p=bit->pcdata_chars; *p; p++)
	character(f, level, dtd, ecw, *p);
}

static void character(FILE16 *f, int level, Dtd dtd, char *ecw, int c)
{
    indent(f, level);
    Fprintf(f, "<character>\n");

    indent(f, level+1);
    Fprintf(f, "<characterCode>%d</characterCode>\n", c);

    simple(f, level+1, "elementContentWhitespace", ecw);

    indent(f, level);
    Fprintf(f, "</character>\n");
}

static void comment(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    indent(f, level);
    Fprintf(f, "<comment>\n");

    Simple(f, level+1, "content", bit->comment_chars);

    indent(f, level);
    Fprintf(f, "</comment>\n");
}

static void attributes(FILE16 *f, int level, Dtd dtd, XBit bit, int *acounter)
{
    Attribute a;

    indent(f, level);
    Fprintf(f, "<attributes>\n");

    for(a=bit->attributes; a; a=a->next)
	if(!a->definition->ns_attr_prefix)
	    attribute(f, level+1, a);

    indent(f, level);
    Fprintf(f, "</attributes>\n");
}

static void namespace_attributes(FILE16 *f, int level, Dtd dtd, XBit bit, int *acounter)
{
    Attribute a;

    indent(f, level);
    Fprintf(f, "<namespaceAttributes>\n");

    for(a=bit->attributes; a; a=a->next)
	if(a->definition->ns_attr_prefix)
	    attribute(f, level+1, a);

    indent(f, level);
    Fprintf(f, "</namespaceAttributes>\n");
}

static void attribute(FILE16 *f, int level, Attribute a)
{
    static Char empty[] = {0};
    char8 *nsname;

    indent(f, level);
    Fprintf(f, "<attribute>\n");

    if(a->ns_definition)
	nsname = a->ns_definition->namespace->uri;
    else if(a->definition->ns_attr_prefix)
	nsname = "http://www.w3.org/2000/xmlns/";
    else
	nsname = 0;

    name(f, level+1,
	 nsname,
	 a->definition->local,
	 a->definition->prefix ? a->definition->prefix : empty);

    Simple(f, level+1, "normalizedValue", a->value);

    simple(f, level+1, "specified", a->specified ? "true" : "false");

    simple(f, level+1, "attributeType",
	   a->definition->declared ? 
	     AttributeTypeName[a->definition->type] : 0);

    indent(f, level);
    Fprintf(f, "</attribute>\n");
}

static void inscope_namespaces(FILE16 *f, int level, Dtd dtd, XBit bit)
{
    NamespaceBinding nsb, nsb2;
    static Char empty[] = {0};

    indent(f, level);
    Fprintf(f, "<inScopeNamespaces>\n");

    for(nsb = bit->ns_dict; nsb; nsb = nsb->parent)
    {
	if(!nsb->namespace)
	    continue;		/* unbinding */

	for(nsb2 = bit->ns_dict; nsb2 != nsb; nsb2 = nsb2->parent)
	    if((!nsb->prefix && !nsb2->prefix) || 
	       (nsb->prefix && nsb2->prefix &&
		Strcmp(nsb->prefix, nsb2->prefix) == 0))
		goto dup;
	
	indent(f, level+1);
	Fprintf(f, "<namespace>\n");

	Simple(f, level+2, "prefix", nsb->prefix ? nsb->prefix : empty);

	simple(f, level+2, "namespaceName", nsb->namespace->uri);

	indent(f, level+1);
	Fprintf(f, "</namespace>\n");
    dup:
	continue;
    }

    indent(f, level);
    Fprintf(f, "</inScopeNamespaces>\n");
}

static void name(FILE16 *f, int level, 
		 const char8 *uri, const Char *local, const Char *prefix)
{
    simple(f, level, "namespaceName", uri);
    Simple(f, level, "localName", local);
    Simple(f, level, "prefix", prefix);
}

static Char *make_id(const char *type, const Char *name, int count)
{
    Char *id = malloc((strlen(type) + 1 + Strlen(name) + 10) * sizeof(Char));
    int i;

    if(count > 0)
	Sprintf(id, InternalCharacterEncoding, "%s-%S-%d", type, name, count);
    else
	Sprintf(id, InternalCharacterEncoding, "%s-%S", type, name);

    /* Change colon to hyphen so it is a namespace-valid ID.  Don't
       worry about uniqueness (eg foo:bar vs foo_bar) since they will
       have different count values anway. */

    for(i=0; id[i]; i++)
	if(id[i] == ':')
	    id[i] = '-';

    return id;
}

static void escape(FILE16 *f, const char8 *text)
{
    while(*text)
	escape1(f, *text++);
}

static void Escape(FILE16 *f, const Char *text)
{
    while(*text)
	escape1(f, *text++);
}

static void escape1(FILE16 *f, int c)
{
    switch(c)
    {
    case '&':
	Fprintf(f, "&amp;");
	break;
    case '<':
	Fprintf(f, "&lt;");
	break;
    case '"':
	Fprintf(f, "&quot;");
	break;
    default:
	Fprintf(f, "%c", c);
	break;
    }
}
